import logging
import os
import tempfile
import zipfile
from enum import Enum, IntEnum
from pathlib import Path
from typing import List

import numpy as np
import requests
from tqdm import tqdm

from tasi.dataset import TrafficLightDataset, TrajectoryDataset

__all__ = [
    "DLRDatasetManager",
    "DLRUTVersion",
    "DLRUTDatasetManager",
    "ObjectClass",
    "DLRTrajectoryDataset",
    "DLRUTTrafficLightDataset",
    "DLRHTVersion",
    "DLRHTDatasetManager",
    "download",
]


class DLRDatasetManager:
    """A base class for DLR dataset management

    Attributes:
        BASE_URI: The base URI for all DLR datasets on zenodo
    """

    BASE_URI: str = "https://zenodo.org/records"
    """str: The base URI of all DLR datasets
    """

    DATA_TYPES = {
        "air_quality": "raw_data",
        "road_condition": "raw_data",
        "traffic_lights": "raw_data",
        "trajectories": "raw_data",
        "weather": "raw_data",
        "traffic_volume": "meta_data",
        "openscenario": "meta_data",
    }
    VERSION_ENUM = None

    @property
    def archivename(self):
        """The base name of the archive"""
        return self.ARCHIVE[self.version]

    @property
    def filename(self):
        """The full name of the archive including the version"""
        return f"{self.archivename}_{self.version.replace('.', '-')}.zip"

    @property
    def url(self):
        """The URL to download the dataset from"""
        return f"{self.BASE_URI}/{self.VERSION[self.version]}/files/{self.filename}"

    @property
    def version(self):
        """The dataset version"""
        return self._version

    @property
    def name(self):

        # fix name of DLR UT v1.0.1 dataset
        if self.version == DLRUTVersion.v1_0_1.value:
            return "DLR-UT_v1-0-0"

        return f"{self.archivename}_{self.version.replace('.', '-')}"

    def __init__(self, version: str, path: str = '/tmp', download_chunk_size: int = 1024, **kwargs):

        if version == 'latest':
            version = self.VERSION_ENUM.latest

        self._version = version.value if isinstance(version, Enum) else version

        self._path = path

        self._chunk_size = download_chunk_size

        super().__init__(**kwargs)

    def load(self, path: Path = None) -> str:
        """
        Download a specified DLR dataset.

        Args:
            path (Path, optional): The destination path where the dataset will be saved.

        Returns:
            str: The path of the exported dataset.
        """

        if path is None:
            path = self._path

        # ensure format of path
        path = path if isinstance(path, Path) else Path(path)

        # define final path
        export_path = path.joinpath(self.name)

        logging.info("Checking if dataset already downloaded %s", export_path.absolute())
        if not export_path.exists():

            logging.info(f"Downloading dataset from {self.url}")

            response = requests.get(self.url, stream=True)
            total_size = int(response.headers.get('content-length', 0))
            with tempfile.NamedTemporaryFile("w+b") as f:

                with tqdm(total=total_size, unit='B', unit_scale=True, desc=f'Downloading {self.name}') as pbar:
                    for chunk in response.iter_content(chunk_size=self._chunk_size):
                        if chunk:
                            f.write(chunk)
                            pbar.update(len(chunk))

                logging.info("Extract files from archive")
                # extract the zip file
                with zipfile.ZipFile(f) as tempzip:
                    file_list = tempzip.namelist()
                    total_files = len(file_list)

                    with tqdm(total=total_files, unit='file', desc='Extracting') as pbar:
                        for file_info in tempzip.infolist():
                            pbar.update(1)
                            tempzip.extract(file_info, path=path.absolute())
        else:
            logging.info(f"Dataset already available at {export_path}")

        return export_path

    def _dataset(self, variant: str, path: Path = None) -> List[str]:
        """Searches for files in the dataset specified at ``path`` for dataset information ``variant``

        Args:
            path (Path): The path of the dataset.
            variant (str): The dataset information to search for

        Returns:
            List[str]: The files found in the dataset for the specified dataset information
        """
        raise NotImplementedError("This method is implemented in child classes.")

    def trajectory(self, path: Path = None) -> List[str]:
        """List of files with trajectory data.

        Args:
            path (Path): The path of the dataset

        Returns:
            List[str]: The files with trajectory data
        """
        return self._dataset("trajectories", path)

    def weather(self, path: Path = None) -> List[str]:
        """List of files with weather data.

        Args:
            path (Path): The path of the dataset

        Returns:
            List[str]: The files with weather data
        """
        return self._dataset("weather", path)

    def road_condition(self, path: Path = None) -> List[str]:
        """List of files with road condition information.

        Args:
            path (Path): The path of the dataset

        Returns:
            List[str]: The files with road condition data
        """
        return self._dataset("road_condition", path)

    def traffic_volume(self, path: Path = None) -> List[str]:
        """List of files with traffic volume data.

        Args:
            path (Path): The path of the dataset

        Returns:
            List[str]: The files with traffic volume data
        """
        return self._dataset("traffic_volume", path)

    def openscenario(self, path: Path = None) -> List[str]:
        """List of files with OpenSCENARIO data.

        Args:
            path (Path): The path of the dataset

        Returns:
            List[str]: The files with OpenSCENARIO data
        """
        return self._dataset("openscenario", path)


class DLRUTVersion(Enum):
    """The available version of the DLR UT dataset"""

    v1_0_0 = "v1.0.0"
    """The initial version of the dataset
    """
    v1_0_1 = "v1.0.1"
    """Contains only minor modifications in the documentation
    """
    v1_1_0 = "v1.1.0"
    """The road condition information was moved into a new sub dataset from the weather data.
    """
    v1_2_0 = "v1.2.0"
    """New folder structure to split raw and metadata. Add new metadata "traffic_volume" and "openscenario". Improve classification of pedestrians and bicycles.
    """
    latest = "v1.2.0"
    """The latest version of the dataset
    """


class DLRUTDatasetManager(DLRDatasetManager):
    """A manager to load the DLR UT dataset from zenodo"""

    VERSION_ENUM = DLRUTVersion
    VERSION = {
        DLRUTVersion.v1_0_0.value: 11396372,
        DLRUTVersion.v1_0_1.value: 13907201,
        DLRUTVersion.v1_1_0.value: 14025010,
        DLRUTVersion.v1_2_0.value: 14773161,
        "latest": 14773161,
    }
    """Dict[str, int]: An internal mapping between version and the zenodo id
    """

    ARCHIVE = dict({
        DLRUTVersion.v1_0_0.value: "DLR-UT"
    },
                   **{
                       v: "DLR-Urban-Traffic-dataset"
                       for v in list(VERSION.keys())[1:]
                   })

    @classmethod
    def area(cls):
        return "urban"

    def _dataset(self, variant: str, path: Path = None) -> List[str]:
        """Searches for files in the dataset specified at ``path`` for dataset information ``variant``

        Args:
            path (Path): The path of the dataset.
            variant (str): The dataset information to search for

        Returns:
            List[str]: The files found in the dataset for the specified dataset information
        """
        if path is None:
            path = self._path

        if not isinstance(path, Path):
            path = Path(path)

        # join dataset path and name
        path = path.joinpath(self.name)

        # add type of data
        if self.version not in [
            DLRUTVersion.v1_0_0.value,
            DLRUTVersion.v1_0_1.value,
            DLRUTVersion.v1_1_0.value,
        ]:
            path = path.joinpath(self.DATA_TYPES[variant])

        # add variant of data
        path = path.joinpath(variant)

        # return file pathes of variant
        return [os.path.join(path, p) for p in sorted(os.listdir(path))]

    def traffic_lights(self, path: Path = None) -> List[str]:
        """List of files with traffic light data.

        Args:
            path (Path): The path of the dataset

        Returns:
            List[str]: The files with traffic light data
        """
        return self._dataset("traffic_lights", path)

    def air_quality(self, path: Path = None) -> List[str]:
        """List of files with air quality data.

        Args:
            path (Path): The path of the dataset

        Returns:
            List[str]: The files with air quality data
        """
        return self._dataset("air_quality", path)


class ObjectClass(IntEnum):
    """
    The supported object classes
    """

    pedestrian = 0
    bicycle = 1
    motorbike = 2
    car = 3
    van = 4
    truck = 5


class DLRTrajectoryDataset(TrajectoryDataset):

    @property
    def pedestrians(self):
        """
        Return the pedestrians of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all pedestrians.
        """
        return self.get_by_object_class(ObjectClass.pedestrian)

    @property
    def bicycles(self):
        """
        Return the bicycles of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all bicycles.
        """
        return self.get_by_object_class(ObjectClass.bicycle)

    @property
    def motorbikes(self):
        """
        Return the motorbikes of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all motorbikes.
        """
        return self.get_by_object_class(ObjectClass.motorbike)

    @property
    def cars(self):
        """
        Return the cars of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all cars.
        """
        return self.get_by_object_class(ObjectClass.car)

    @property
    def vans(self):
        """
        Return the vans of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all vans.
        """
        return self.get_by_object_class(ObjectClass.van)

    @property
    def trucks(self):
        """
        Return the trucks of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all trucks.
        """
        return self.get_by_object_class(ObjectClass.truck)

    @property
    def mru(self):
        """
        Return the motorized road user of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all motorized objects.
        """
        return self.get_by_object_class([ObjectClass.motorbike, ObjectClass.car, ObjectClass.van, ObjectClass.truck])

    @property
    def vru(self):
        """
        Return the vulnerable road user of the dataset.

        Returns:
            DLRTrajectoryDataset: Dataset of all motorized objects.
        """
        return self.get_by_object_class([ObjectClass.pedestrian, ObjectClass.bicycle])

    @property
    def roi(self):
        """
        Return the region of interest of the dataset.

        Returns:
            np.ndarray: The region of interest.
        """
        return np.array([
            self.center.easting.min(),
            self.center.northing.min(),
            self.center.easting.max(),
            self.center.northing.max(),
        ]).reshape(-1, 2)


class DLRUTTrafficLightDataset(TrafficLightDataset):

    def signal(self, signal_id: int):
        """
        Filter the dataset by a signal id.

        Args:
            signal_id (int): The id of the signal.

        Returns:
            `DLRUTTrafficLightDataset`: The data from the signal
        """
        return self.xs(signal_id, level=1)

    def signal_state(self, signal_state: int):
        """
        Filter the dataset by an signal state.

        Args:
            signal_state (int): The signal state used for filtering.

        Returns:
            `DLRUTTrafficLightDataset`: The data with the user defined signal state.
        """
        return self.loc[self["state"] == signal_state]


class DLRHTVersion(Enum):
    """The available version of the DLR HT dataset"""

    v1_0_0 = "v1.0.0"
    """The initial version of the dataset
    """
    v1_1_0 = "v1.1.0"
    """Add new metadata "openscenario" and improve object classification.
    """
    latest = "v1.1.0"
    """The latest version of the dataset
    """


class DLRHTDatasetManager(DLRDatasetManager):
    """A manager to load the DLR HT dataset from zenodo"""

    VERSION_ENUM = DLRHTVersion
    VERSION = {
        DLRHTVersion.v1_0_0.value: 14012006,
        DLRHTVersion.v1_1_0.value: 14811064,
        "latest": 14811064,
    }
    """Dict[str, int]: An internal mapping between version and the zenodo id
    """

    @classmethod
    def area(cls):
        return "highway"

    @property
    def archivename(self):
        """The base name of the archive"""
        return "DLR-Highway-Traffic-dataset"

    def _dataset(self, variant: str, path: Path = None) -> List[str]:
        """Searches for files in the dataset specified at ``path`` for dataset information ``variant``

        Args:
            path (Path): The path of the dataset.
            variant (str): The dataset information to search for

        Returns:
            List[str]: The files found in the dataset for the specified dataset information
        """
        if path is None:
            path = self._path

        if not isinstance(path, Path):
            path = Path(path)

        # join dataset path and name, type and variant
        path = (path.joinpath(self.name).joinpath(self.DATA_TYPES[variant]).joinpath(variant))

        # return file pathes of variant
        return [os.path.join(path, p) for p in sorted(os.listdir(path))]


def download():

    from tasi.logging import init_logger

    init_logger()

    import argparse
    import sys

    parser = argparse.ArgumentParser(prog="dlr-downloader")
    parser.add_argument("--name", choices=["urban", "highway"], type=str, required=True)
    parser.add_argument("--version", type=str, required=True)
    parser.add_argument("--path", type=str, required=True)
    arguments = parser.parse_args(sys.argv[1:])

    # choose required dataset
    if arguments.name.lower() == "urban":
        dataset_cls = DLRUTDatasetManager
    elif arguments.name.lower() == "highway":
        dataset_cls = DLRHTDatasetManager

    # ensure valid format of version
    version = arguments.version.replace("-", ".").replace("_", ".")
    if not version.startswith("v") and version != "latest":
        version = "v" + version

    dataset = dataset_cls(version=version)
    dataset.load(path=Path(arguments.path))


if __name__ == "__main__":

    download()



from datetime import datetime
from typing import Iterable, List, Union

import numpy as np
import pandas as pd
from typing_extensions import Self

from tasi.utils import to_pandas_multiindex

__all__ = [
    "Dataset",
    "TrajectoryDataset",
    "WeatherDataset",
    "AirQualityDataset",
    "RoadConditionDataset",
    "TrafficLightDataset",
    "TrafficVolumeDataset",
]

ObjectClass = Union[str, int]


class Dataset(pd.DataFrame):

    TIMESTAMP_COLUMN = "timestamp"
    ID_COLUMN = "id"
    INDEX_COLUMNS = [TIMESTAMP_COLUMN, ID_COLUMN]

    @property
    def _constructor(self):
        return self.__class__

    @property
    def timestamps(self) -> np.ndarray[np.datetime64]:
        """The unique timestamps in the dataset

        Returns:
            np.ndarray[np.datetime64]: A list of timestamps
        """
        return self.index.get_level_values(self.TIMESTAMP_COLUMN).unique()

    @property
    def ids(self) -> np.ndarray[np.int64]:
        """Returns the unique ids in the dataset

        Returns:
            np.ndarray: A List of ids
        """
        try:
            return self.index.get_level_values(self.ID_COLUMN).unique()
        except BaseException:
            return self[self.ID_COLUMN].unique()

    @property
    def attributes(self) -> np.ndarray[str]:
        """Returns the dataset attributes

        Returns:
            np.ndarray: A list of attribute names
        """
        return self.columns.get_level_values(0).unique()

    def att(
        self,
        timestamps: Union[pd.Timestamp, List[pd.Timestamp], pd.Index],
        attribute: Union[List[str], pd.Index] = None,
    ) -> Union[Self, pd.Series]:
        """Select the rows at the specified times and optionally the specified attributes.

        Args:
            timestamps (Union[pd.Timestamp, List[pd.Timestamp], pd.Index]): The timestamps to select
            attribute (pd.Index, optional): The attribute to select. Defaults to None.

        Returns:
            Union[Self, pd.Series]: The selected row(s) and column(s)
        """
        try:
            len(timestamps)
        except BaseException:
            timestamps = [timestamps]

        if attribute is None:
            return self.loc[pd.IndexSlice[timestamps, :]]
        else:
            return self.loc[pd.IndexSlice[timestamps, :], attribute]

    def atid(
        self, ids: Union[int, List[int], pd.Index], attributes: pd.Index = None
    ) -> Self:
        """Select rows by the given id and optionally by attributes

        Args:
            ids (Union[int, List[int], pd.Index]): A list of IDs
            attributes (pd.Index, optional): A list of attribute names. Defaults to None.

        Returns:
            Self: The selected rows and attributes
        """
        try:
            len(ids)
        except BaseException:
            ids = [ids]

        if attributes is None:
            return self.loc[pd.IndexSlice[:, ids], :]
        else:
            return self.loc[pd.IndexSlice[:, ids], attributes]

    @property
    def interval(self) -> pd.Interval:
        """Returns the time interval this dataset spans

        Returns:
            pd.Interval: The interval of this
        """
        return pd.Interval(self.timestamps[0], self.timestamps[-1])

    def during(self, since: datetime, until: datetime, include_until: bool = False):
        """
        Select rows within a specific time range (include "since", exclude "until").

        Args:
            since (datetime): The start datetime for the selection.
            until (datetime): The end datetime for the selection.
            include_until (bool, optional): Whether to include data with timestamp "until". Defaults to False.

        Returns:
            ObjectDataset: A subset of the dataset with rows between the specified datetimes.
        """

        # get all timestamps
        timestamps = self.index.get_level_values(self.TIMESTAMP_COLUMN)

        # create a mask selecting only the relevant point in times
        valid_since = timestamps >= since
        if include_until:
            valid_until = timestamps <= until
        else:
            valid_until = timestamps < until

        # select the entries
        return self.loc[valid_since & valid_until]

    @classmethod
    def from_csv(cls, file: str, indices: Union[List, str] = (), **kwargs) -> Self:
        """
        Read a dictionary-alike object from a `.csv` file as a pandas DataFrame.

        Args:
            file (str): The path and name of the dataset `.csv` file
            indices (Union[List, str]): The name of the columns to use as index

        """
        if indices and not hasattr(kwargs, "index_col"):
            kwargs["index_col"] = indices

        # read csv data
        df = pd.read_csv(file, **kwargs)

        # parse dates
        df["timestamp"] = pd.to_datetime(df["timestamp"], format="ISO8601")

        # try to set index
        try:
            df.set_index(cls.INDEX_COLUMNS, inplace=True)
        except KeyError:
            try:
                df.set_index(cls.TIMESTAMP_COLUMN, inplace=True)
            except KeyError:
                pass

        return cls(df)


class TrajectoryDataset(Dataset):

    def trajectory(self, index: Union[int, Iterable[int]], inverse: bool = False):
        """
        Select trajectory data for specific indices, or exclude them if inverse is set to True.

        Args:
            index (Union[int, Iterable[int]], optional): An integer or an iterable of integers representing the indices of
                the trajectories to select. If a single integer is provided, only the trajectory corresponding
                to that index is selected. If a list or other iterable of integers is provided, all trajectories
                corresponding to those indices are selected.
            inverse (bool, optional): If set to True, the selection is inverted, meaning the specified indices
                are excluded from the resulting dataset, and all other trajectories are included. Defaults to False.

        Returns:
            TrajectoryDataset: A trajectory or multiple trajectories of the dataset.
        """

        if isinstance(index, int):
            index = [index]

        if inverse:
            index = self.ids.difference(index)

        return self.atid(index)

    def most_likely_class(
        self, by: str = "trajectory", broadcast: bool = False
    ) -> pd.Series:
        """
        Get the name of the most probable object class for each pose or trajectory of the dataset

        Args:
            by (str): By which object the most likely class should be determined. Possible values: 'pose', 'trajectory'
            broadcast (bool, optional): Specifies whether the most likely class should be broadcasted to each pose of the dataset.
                The option only changes the output for trajectories. Defaults to False.

        Returns:
            pd.Series: Information about the most probable object class.
                If `by` is "pose" and broadcast is "False" or "True" return the most likely object class of each pose.
                If `by` is "trajectory" and broadcast is "False" return the most likely object class of each trajectory.
                If `by` is "trajectory" and broadcast is "True" return the most likely object class of each trajectory
                for each pose.

        Raises:
            ValueError: If the value of "by" is neither 'pose' nor 'trajectory'.
        """
        if by == "pose":
            return self.classifications.idxmax(axis=1)

        elif by == "trajectory":
            trajectory_class = self.classifications.groupby("id").apply(
                lambda tj_classes: tj_classes.mean().idxmax()
            )

            if broadcast:
                return self.apply(lambda ser: trajectory_class[ser.name[1]], axis=1)
            else:
                return trajectory_class

        else:
            raise ValueError("'by' must be one of 'pose' or 'trajectory'.")

    def get_by_object_class(self, object_class: Union[List[ObjectClass], ObjectClass]):
        """
        Return only the poses of a specific object class.

        Args:
            object_class (ObjectClass): The object class.

        Returns:
            ObjectDataset: Dataset containing only the poses of a defined object class.

        Note:
            The object class of a pose is determined by the mean probability of all poses in the trajectory.
        """

        if not isinstance(object_class, list):
            object_class = [object_class]

        return self[
            self.most_likely_class(by="trajectory", broadcast=True).isin(object_class)
        ]

    @classmethod
    def from_csv(cls, *args, **kwargs) -> Self:

        df = super().from_csv(*args, **kwargs)

        # ensure the column is a pandas MultiIndex
        df.columns = to_pandas_multiindex(df.columns.to_list())

        return cls(df)


class WeatherDataset(Dataset):

    pass


class AirQualityDataset(Dataset):

    pass


class RoadConditionDataset(Dataset):

    pass


class TrafficLightDataset(Dataset):

    pass


class TrafficVolumeDataset(Dataset):

    @classmethod
    def from_csv(cls, *args, **kwargs) -> Self:

        df = super().from_csv(*args, **kwargs)

        # transform data to Dataset format
        df = df.stack().to_frame("volume")

        # ensure index names are correnct
        df.index.names = cls.INDEX_COLUMNS

        return cls(df)

    @property
    def lanes(self) -> pd.Index:
        """Returns the unique lanes of the dataset

        Returns:
            pd.Index: The unique lanes as Index
        """
        return self.ids

    def lane(self, lane: str) -> pd.Series:
        """Returns the traffic volume from a specific lane

        Returns:
            pd.Series: A pd.Series of traffic volume data
        """
        return self.atid(lane).volume



from typing import List
import numpy as np
import pandas as pd
from copy import copy


def to_pandas_multiindex(values: List[str], separator="_") -> pd.MultiIndex:
    """
    Convert a list of nested names to a `pandas.MultiIndex`.

    Args:
        values (List[str]): The list of names joined by the `separator`
        separator (str, optional): The character used as separator. Defaults to '_'.

    Returns:
        pd.MultiIndex: The list of nested names converted to a `pandas.MultiIndex`

    Examples:
        This function can be employed if you have a list of attributes names where the attributes are hierarchical.
        Let's assume we have the attribute `a` with the components `b` and `c` and the attribute `v` with the components
        `x` and `y` which are joined by the separator `|` so that there is a list with the values `['a|b', 'a|c', 'v|x',
        'v|y']` we want to convert to a `pandas.MultiIndex` since this is used to manage hierarchical attributes.
        Therefore, we can use this function with::

            >>> attributes = ['a|b', 'a|c', 'v|x', 'v|y']
            >>> to_pandas_multiindex(attributes, separator='|')
            MultiIndex([('a', 'b'),
                        ('a', 'c'),
                        ('v', 'x'),
                        ('v', 'y')],
                       )

    """

    columns = copy(values)

    # get the maximum number of multiindex rows
    depth = max([c.count(separator) for c in columns]) + 1

    names = []
    # for each attribute get the corresponding columns
    for attr in values:
        values = attr.split(separator)

        # get the missing inner names
        missing_names = depth - len(values)

        # append the missing names
        if missing_names > 0:
            values.extend([""] * missing_names)

        names.append(values)

    # to set the name of the columns with a multiindex we need to transpose the list of columns names from a
    # column-wise to a row-wise representation
    return pd.MultiIndex.from_arrays(np.array(names).T)



from typing import ClassVar, Tuple
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap
import numpy as np
from tilemapbase.mapping import Plotter

from tasi.dataset import TrajectoryDataset
from tasi.plotting.wms import BoundingboxTiles


class BoundingboxPlotter(Plotter):
    """
    A specialization of the `tilemapbase.Plotter` to show tiles given a region as a boundingbox in UTM coordinates.
    """

    def __init__(
        self, extent: np.ndarray, tile_provider: BoundingboxTiles, padding: int = 0
    ):
        """
        Create a plotter for the given `extend` and using the provided `tile_provider`

        Args:
            extent (np.ndarray): A 2*2 matrix as the 2-point definition of the region to plot.
            tile_provider (ClassVar[BoundingboxTiles]): The tile provider to use for fetching tiles
            padding (int, optional): An additional padding in meters around the given extend. Defaults to 0.
        """

        self._extent = extent
        self._original_extent = extent
        self._tile_provider = tile_provider

        self._padding = padding

    def plot(
        self,
        ax: plt.Axes,
        center: Tuple[float, float] = None,
        zoom: float = 1,
        show_attribution: bool = True,
        attribution_kwargs=None,
        **kwargs
    ):
        """
        Draw the tile for the center location given by ``center`` within the current extend and and zoom into the tile according to `zoom`.

        Args:
            ax (plt.Axes): The axes to plot onto.
            center (Tuple[float, float]): The plotting center
            zoom (float): The zoom value
            show_attribution (bool): To thow the attribution information. Defaults to True.

        Raises:
            ValueError: If the zoom value is not within (0,1).

        """

        if zoom == 0 or zoom > 1:
            raise ValueError("The zoom value needs to be within (0,1).")

        # get the tile for the current extend
        tile = self._tile_provider.get_tile(
            self.xtilemin, self.ytilemin, self.xtilemax, self.ytilemax
        )

        # get the size of the extend
        dx = self.xtilemax - self.xtilemin
        dy = self.ytilemax - self.ytilemin

        # draw the tile on the axes and specify the extend in the given coordinate system
        ax.imshow(
            tile,
            interpolation="lanczos",
            extent=[
                int(self.xtilemin),
                int(self.xtilemax),
                int(self.ytilemin),
                int(self.ytilemax),
            ],
            **kwargs,
        )
        if center is None:
            center = [self.xtilemin + (dx / 2), self.ytilemin + (dy / 2)]

        # set the limit of the axes according to the given extend and zoom value
        ax.set(
            xlim=[center[0] - (dx / 2) * zoom, center[0] + (dx / 2) * zoom],
            ylim=[center[1] - (dy / 2) * zoom, center[1] + (dy / 2) * zoom],
        )

        if show_attribution:

            default_config = {
                "fontsize": 5,
                "transform": ax.transAxes,
                "s": self._tile_provider.ATTRIBUTION,
            }

            attribution_kwargs = (
                attribution_kwargs if attribution_kwargs is not None else {}
            )

            if dx > dy:

                default_config.update(
                    {"x": 1, "y": 1.01, "ha": "right", "va": "bottom"}
                )

                default_config.update(attribution_kwargs)

                ax.text(**default_config)

            else:

                default_config.update(
                    {"x": 1.02, "y": 0.01, "rotation": 90,
                        "ha": "left", "va": "bottom"}
                )

                default_config.update(attribution_kwargs)

                ax.text(**default_config)
    @property
    def extent(self):
        """
        The region in 2-point definition

        Returns:
            np.ndarray: A 2*2 matrix
        """
        return self._original_extent

    @property
    def xtilemin(self) -> int:
        """
        The smallest x-coordinate in tilespace

        Returns:
            int: Minimum easting
        """
        return int(np.min(self._extent[:, 0])) - self._padding

    @property
    def xtilemax(self) -> int:
        """
        The greatest x-coordinate in tilespace

        Returns:
            int: Maximum easting
        """
        return int(np.max(self._extent[:, 0])) + self._padding

    @property
    def ytilemin(self) -> int:
        """
        The smallest y-coordinate in tilespace

        Returns:
            int: Minimum northing
        """
        return int(np.min(self._extent[:, 1])) - self._padding

    @property
    def ytilemax(self) -> int:
        """
        The greatest y-coordinate in tilespace

        Returns:
            int: Maximum northing
        """
        return int(np.max(self._extent[:, 1])) + self._padding


class TrajectoryPlotter:
    """
    Plot trajectories using ``matplotlib``
    """

    OBJECT_CLASS_COLORS = None

    def __init__(self, color_palette: str = "tab10"):

        self.palette = plt.cm.get_cmap(color_palette)

        if not isinstance(self.palette, ListedColormap):
            raise TypeError(
                "Argument 'color_palette' is not of type 'matplotlib.colors.ListedColormap'"
            )

    def plot(
        self,
        dataset: TrajectoryDataset,
        color: str = None,
        ax: Axes = None,
        trajectory_kwargs=None,
        **kwargs
    ):
        """
        Plot trajectories using `matplotlib`

        Args:
            dataset (TrajectoryDataset): The dataset of trajectories to visualize.
            color (str, optional): The color of the trajectories. Defaults to None.
            ax (Axes, optional): The matplotlib axes. Defaults to None.
            trajectory_kwargs (Dict, optional): A mapping of traffic participant id to trajectory-specific plotting attributes. Defaults to None.
        """
        if ax is None:
            ax = plt.gca()

        trajectory_kwargs = trajectory_kwargs if trajectory_kwargs is not None else {}

        # get the classes for each trajectory
        tj_classes = dataset.most_likely_class()

        # get mapping of classes to colors
        if self.OBJECT_CLASS_COLORS is None:
            object_class_colors = {c: i for i,
                                   c in enumerate(tj_classes.unique())}
        else:
            object_class_colors = self.OBJECT_CLASS_COLORS

        for tj_class, tj_id in zip(tj_classes, dataset.ids):

            # get the trajectory of the id
            tj = dataset.trajectory(tj_id)

            # get additional plotting arguments of this trajectory
            tj_kwargs = trajectory_kwargs.get(tj_id, {})

            if "c" not in tj_kwargs and "color" not in tj_kwargs:

                # color for this trajectory is not defined
                if color is None:

                    # set default color
                    tj_kwargs["color"] = self.palette(
                        object_class_colors[tj_class])

                else:

                    # use color of all trajectories
                    tj_kwargs["color"] = color

            if "label" not in tj_kwargs:
                tj_kwargs["label"] = tj_class

            # use the center position for plotting
            ax.plot(
                tj.center.easting,
                tj.center.northing,
                **tj_kwargs,
                **kwargs,
            )

        if len(trajectory_kwargs) == 0:

            # all object classes have the same color -> show legend of unique labels
            handles, labels = plt.gca().get_legend_handles_labels()
            by_label = dict(zip(labels, handles))
            plt.legend(by_label.values(), by_label.keys())



from tasi.plotting.plot import TrajectoryPlotter
from tasi.dlr.dataset import ObjectClass


class DLRTrajectoryPlotter(TrajectoryPlotter):
    """
    Plot DLR trajectories using ``matplotlib``
    """

    OBJECT_CLASS_COLORS = {i.name: i.value for i in ObjectClass}



from .dataset import (
    ObjectClass,
    DLRDatasetManager,
    DLRUTDatasetManager,
    DLRHTDatasetManager,
    DLRTrajectoryDataset,
    DLRUTTrafficLightDataset,
    DLRUTVersion,
    DLRHTVersion,
)



import logging
import sys
from logging.handlers import RotatingFileHandler

import coloredlogs

MSG_TEMPLATE = (
    "[%(asctime)s | %(filename)s:%(funcName)s:%(lineno)d] > %(levelname)s:  %(message)s"
)

LOGGER = "tasi"


def init_logger(
    console_level: int = logging.INFO,
    file_level: int = logging.DEBUG,
    filename: str = None,
):

    # get the current root logger
    logger = logging.getLogger()
    list(map(logger.removeHandler,  logger.handlers))

    # create a format for the logging. For file logging as well as for logging to console
    log_format = logging.Formatter(MSG_TEMPLATE.format(start="", end=""))

    # set the default log level to the console level
    logger.setLevel(console_level)

    if filename:
        # setup a rotating file logger
        file_handler = RotatingFileHandler(filename, pow(2, 20), 10, delay=True)
        file_handler.setFormatter(log_format)
        file_handler.setLevel(file_level)
        logger.addHandler(file_handler)

    # setup the console logging - log to stdout
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(coloredlogs.ColoredFormatter(fmt=MSG_TEMPLATE))
    handler.setLevel(console_level)
    logger.addHandler(handler)



import io as _io
import urllib
from copy import copy

import PIL.Image as _Image
from tilemapbase.tiles import Tiles


class BoundingboxTiles(Tiles):
    """
    A base class to provide tiles from an URL given a region as a boundingbox
    """

    DEFAULT_PARAMS = dict(
        width=512,
        height=512,
        crs="EPSG:25833",
        format="image/png",
        request="GetMap",
        layers="",
        version="",
        styles="",
    )
    """Dict: The default parameter to query a WMS server (GET parameters)"""

    WMS = None
    """str: The WMS URL to query"""

    SOURCE_NAME = None
    """str: The source name"""

    ATTRIBUTION = ""
    """str: Attribution to the WMS layer"""

    def __init__(self, width: float = None, height: float = None, *args, **kwargs):

        self.params = copy(self.DEFAULT_PARAMS)

        for key, value in kwargs.items():
            self.params[key] = value

        self._width = self.params.get("width", None) if width is None else width
        self._height = self.params.get("height", None) if height is None else height

        for attr in ["width", "height"]:
            if attr in self.params:
                del self.params[attr]

        if self.width is None and self.height is None:
            raise ValueError("Either specify a tile width or height")

        # This is a mandatory argument to the `Tiles` class, though we don't use it.
        kwargs["request_string"] = ""
        kwargs["source_name"] = self.SOURCE_NAME

        super().__init__(*args, **kwargs)

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    def _request_string(self, x1, y1, x2, y2):
        """Encodes the tile coords and name into a string for the query."""

        dx = x2 - x1
        dy = y2 - y1

        width = self.width
        height = self.height

        if width is not None:
            height = int(width / dx * dy)
        elif height is not None:
            width = int(height / dy * dx)

        return (
            f"{self.WMS.format(XMIN=int(x1), YMIN=int(y1), XMAX=int(x2), YMAX=int(y2))}"
            f"&{urllib.parse.urlencode({**self.params, **dict(width=width, height=height)})}"
        )

    def _request_http(self, request_string):
        return request_string

    def get_tile(self, x1, y1, x2, y2):
        tile = self._get_cache().fetch(self._request_string(x1, y1, x2, y2))

        if tile is None:
            return None
        try:
            fp = _io.BytesIO(tile)
            return _Image.open(fp)
        except BaseException:
            raise RuntimeError(
                "Failed to decode data for {} - @ {} extent".format(
                    self.name, [x1, y1, x2, y2]
                )
            )


class LowerSaxonyOrthophotoTile(BoundingboxTiles):
    """Tile that provides access to the WMS server of the LGLN."""

    WMS = "https://opendata.lgln.niedersachsen.de/doorman/noauth/dop_wms?bbox={XMIN},{YMIN},{XMAX},{YMAX}"

    DEFAULT_PARAMS = dict(
        width=512,
        height=512,
        service="WMS",
        crs="EPSG:25832",
        format="image/png",
        request="GetMap",
        layers="ni_dop20",
        styles="",
        version="1.3.0",
    )

    SOURCE_NAME = "LGLN"

    ATTRIBUTION = "(C) GeoBasis-DE/LGLN 2024 CC-BY 4.0"
